from dataclasses import dataclass
from enum import Enum, unique


@unique
class RelType(Enum):
    COMPOSITION = '*'
    INHERITANCE = '<|'


@dataclass
class UmlRelation:
    source_fqn: str
    target_fqn: str
    type: RelType



from typing import Dict, Type

from py2puml.domain.umlclass import UmlAttribute, UmlClass
from py2puml.domain.umlitem import UmlItem


def inspect_namedtuple_type(namedtuple_type: Type, namedtuple_type_fqn: str, domain_items_by_fqn: Dict[str, UmlItem]):
    domain_items_by_fqn[namedtuple_type_fqn] = UmlClass(
        name=namedtuple_type.__name__,
        fqn=namedtuple_type_fqn,
        attributes=[UmlAttribute(tuple_field, 'Any', False) for tuple_field in namedtuple_type._fields],
    )



from re import Pattern
from re import compile as re_compile
from typing import Tuple

# a class name wrapped by ForwardRef(...)
FORWARD_REFERENCES: Pattern = re_compile(r"ForwardRef\('([^']+)'\)")

# valid characters for class names and compound types
IS_COMPOUND_TYPE: Pattern = re_compile(r'^[a-z|A-Z|0-9|\[|\]|\.|,|\s|_|\|]+$')

# characters involved in the build-up of compound types
SPLITTING_CHARACTERS = '[', ']', ',', '|'

# 'None' in 'Union[str, None]' type signature is changed into 'NoneType' when inspecting a module
LAST_NONETYPE_IN_UNION: Pattern = re_compile(r'Union\[(?:(?:[^\[\]])*NoneType)')


def remove_forward_references(compound_type_annotation: str, module_name: str) -> str:
    """
    Removes the forward reference mention from the string representation of a type annotation.
    This happens when a class attribute refers to the class being defined (where Person.friends is of type List[Person])
    The type which is referred to is prefixed by the module where it is defined to help resolution.
    """
    return (
        None
        if compound_type_annotation is None
        else FORWARD_REFERENCES.sub(f'{module_name}.\\1', compound_type_annotation)
    )


def replace_nonetype_occurrences_in_union_types(type_annotation: str) -> str:
    """
    `None` types are replaced by `NoneType` during code inspection in type annotations like `Union[str, None]`.
    This function replaces back the `NoneType` occurrences to `None`
    See https://bugs.python.org/issue44635
    """
    if type_annotation is None:
        return None

    cleaned_type_annotation = type_annotation

    while (union_match_clauses := list(LAST_NONETYPE_IN_UNION.finditer(cleaned_type_annotation))) and len(
        union_match_clauses
    ) > 0:
        union_last_match_clause = union_match_clauses[-1]
        match_start, match_end = union_last_match_clause.span()
        cleaned_type_annotation = (
            cleaned_type_annotation[0 : match_end - 8] + 'None' + cleaned_type_annotation[match_end:]
        )

    return cleaned_type_annotation


class CompoundTypeSplitter:
    """
    Splits the representation of a compound type annotation into a list of:
    - its components (that can be resolved against the module where the type annotation was found)
    - its structuring characters: '[', ']' and ','
    """

    def __init__(self, compound_type_annotation: str, module_name: str):
        resolved_type_annotations = remove_forward_references(compound_type_annotation, module_name)
        resolved_type_annotations = replace_nonetype_occurrences_in_union_types(resolved_type_annotations)
        if (resolved_type_annotations is None) or not IS_COMPOUND_TYPE.match(resolved_type_annotations):
            raise ValueError(f'{compound_type_annotation} seems to be an invalid type annotation')

        self.compound_type_annotation = resolved_type_annotations

    def get_parts(self) -> Tuple[str]:
        "Iteratively splits the type annotation with the different SPLITTING_CHARACTERS"

        parts = [self.compound_type_annotation]
        for splitting_character in SPLITTING_CHARACTERS:
            new_parts = []
            for part in parts:
                splitted_parts = part.split(splitting_character)
                new_parts.append(splitted_parts[0])
                # some splitting characters (like ',' and '|' and '[') separate a type annotation into different types
                # others (like ']') must just be separated from the text
                if len(splitted_parts) > 1:
                    for splitted_part in splitted_parts[1:]:
                        new_parts.extend([splitting_character, splitted_part])
            parts = (new_part.strip() for new_part in new_parts if len(new_part.strip()) > 0)

        return tuple(parts)



from dataclasses import is_dataclass
from enum import Enum
from inspect import getmembers, isclass
from types import ModuleType
from typing import Dict, Iterable, List, Type

from py2puml.domain.umlitem import UmlItem
from py2puml.domain.umlrelation import UmlRelation
from py2puml.inspection.inspectclass import inspect_class_type, inspect_dataclass_type
from py2puml.inspection.inspectenum import inspect_enum_type
from py2puml.inspection.inspectnamedtuple import inspect_namedtuple_type


def filter_domain_definitions(module: ModuleType, root_module_name: str) -> Iterable[Type]:
    for definition_key in dir(module):
        definition_type = getattr(module, definition_key)
        if isclass(definition_type):
            definition_members = getmembers(definition_type)
            definition_module_member = next(
                (
                    member
                    for member in definition_members
                    # ensures that the type belongs to the module being parsed
                    if member[0] == '__module__' and member[1].startswith(root_module_name)
                ),
                None,
            )
            if definition_module_member is not None:
                yield definition_type


def inspect_domain_definition(
    definition_type: Type,
    root_module_name: str,
    domain_items_by_fqn: Dict[str, UmlItem],
    domain_relations: List[UmlRelation],
):
    definition_type_fqn = f'{definition_type.__module__}.{definition_type.__name__}'
    if definition_type_fqn not in domain_items_by_fqn:
        if issubclass(definition_type, Enum):
            inspect_enum_type(definition_type, definition_type_fqn, domain_items_by_fqn)
        elif getattr(definition_type, '_fields', None) is not None:
            inspect_namedtuple_type(definition_type, definition_type_fqn, domain_items_by_fqn)
        elif is_dataclass(definition_type):
            inspect_dataclass_type(
                definition_type, definition_type_fqn, root_module_name, domain_items_by_fqn, domain_relations
            )
        else:
            inspect_class_type(
                definition_type, definition_type_fqn, root_module_name, domain_items_by_fqn, domain_relations
            )


def inspect_module(
    domain_item_module: ModuleType,
    root_module_name: str,
    domain_items_by_fqn: Dict[str, UmlItem],
    domain_relations: List[UmlRelation],
):
    # processes only the definitions declared or imported within the given root module
    for definition_type in filter_domain_definitions(domain_item_module, root_module_name):
        inspect_domain_definition(definition_type, root_module_name, domain_items_by_fqn, domain_relations)



from dataclasses import dataclass
from importlib import import_module
from inspect import isabstract
from re import compile as re_compile
from typing import Dict, List, Type

from py2puml.domain.umlclass import UmlAttribute, UmlClass
from py2puml.domain.umlitem import UmlItem
from py2puml.domain.umlrelation import RelType, UmlRelation
from py2puml.parsing.astvisitors import shorten_compound_type_annotation
from py2puml.parsing.moduleresolver import ModuleResolver
from py2puml.parsing.parseclassconstructor import parse_class_constructor

# from py2puml.utils import investigate_domain_definition

CONCRETE_TYPE_PATTERN = re_compile("^<(?:class|enum) '([\\.|\\w]+)'>$")


def handle_inheritance_relation(
    class_type: Type, class_fqn: str, root_module_name: str, domain_relations: List[UmlRelation]
):
    for base_type in getattr(class_type, '__bases__', ()):
        base_type_fqn = f'{base_type.__module__}.{base_type.__name__}'
        if base_type_fqn.startswith(root_module_name):
            domain_relations.append(UmlRelation(base_type_fqn, class_fqn, RelType.INHERITANCE))


def inspect_static_attributes(
    class_type: Type,
    class_type_fqn: str,
    root_module_name: str,
    domain_items_by_fqn: Dict[str, UmlItem],
    domain_relations: List[UmlRelation],
) -> List[UmlAttribute]:
    """
    Adds the definitions:
    - of the inspected type
    - of its static attributes from the class annotations (type and relation)
    """
    # defines the class being inspected
    definition_attrs: List[UmlAttribute] = []
    uml_class = UmlClass(
        name=class_type.__name__, fqn=class_type_fqn, attributes=definition_attrs, is_abstract=isabstract(class_type)
    )
    domain_items_by_fqn[class_type_fqn] = uml_class
    # investigate_domain_definition(class_type)

    type_annotations = getattr(class_type, '__annotations__', None)
    if type_annotations is not None:
        # stores only once the compositions towards the same class
        relations_by_target_fqdn: Dict[str:UmlRelation] = {}
        # utility which outputs the fully-qualified name of the attribute types
        module_resolver = ModuleResolver(import_module(class_type.__module__))

        # builds the definitions of the class attrbutes and their relationships by iterating over the type annotations
        for attr_name, attr_class in type_annotations.items():
            attr_raw_type = str(attr_class)
            concrete_type_match = CONCRETE_TYPE_PATTERN.search(attr_raw_type)
            # basic type
            if concrete_type_match:
                concrete_type = concrete_type_match.group(1)
                # appends a composition relationship if the attribute is a class from the inspected domain
                if attr_class.__module__.startswith(root_module_name):
                    attr_type = attr_class.__name__
                    attr_fqn = f'{attr_class.__module__}.{attr_class.__name__}'
                    relations_by_target_fqdn[attr_fqn] = UmlRelation(uml_class.fqn, attr_fqn, RelType.COMPOSITION)
                else:
                    attr_type = concrete_type
            # compound type (tuples, lists, dictionaries, etc.)
            else:
                attr_type, full_namespaced_definitions = shorten_compound_type_annotation(
                    attr_raw_type, module_resolver
                )
                relations_by_target_fqdn.update(
                    {
                        attr_fqn: UmlRelation(uml_class.fqn, attr_fqn, RelType.COMPOSITION)
                        for attr_fqn in full_namespaced_definitions
                        if attr_fqn.startswith(root_module_name)
                    }
                )

            uml_attr = UmlAttribute(attr_name, attr_type, static=True)
            definition_attrs.append(uml_attr)

        domain_relations.extend(relations_by_target_fqdn.values())

    return definition_attrs


def inspect_class_type(
    class_type: Type,
    class_type_fqn: str,
    root_module_name: str,
    domain_items_by_fqn: Dict[str, UmlItem],
    domain_relations: List[UmlRelation],
):
    attributes = inspect_static_attributes(
        class_type, class_type_fqn, root_module_name, domain_items_by_fqn, domain_relations
    )
    instance_attributes, compositions = parse_class_constructor(class_type, class_type_fqn, root_module_name)
    attributes.extend(instance_attributes)
    domain_relations.extend(compositions.values())

    handle_inheritance_relation(class_type, class_type_fqn, root_module_name, domain_relations)


def inspect_dataclass_type(
    class_type: Type[dataclass],
    class_type_fqn: str,
    root_module_name: str,
    domain_items_by_fqn: Dict[str, UmlItem],
    domain_relations: List[UmlRelation],
):
    for attribute in inspect_static_attributes(
        class_type, class_type_fqn, root_module_name, domain_items_by_fqn, domain_relations
    ):
        attribute.static = False

    handle_inheritance_relation(class_type, class_type_fqn, root_module_name, domain_relations)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from argparse import ArgumentParser
from pathlib import Path
from sys import path

from py2puml.py2puml import py2puml


def run():
    # adds the current working directory to the system path in the first place
    # to ease module resolution when py2puml imports them
    current_working_directory = str(Path.cwd().resolve())
    path.insert(0, current_working_directory)

    argparser = ArgumentParser(description='Generate PlantUML class diagrams to document your Python application.')

    argparser.add_argument('-v', '--version', action='version', version='py2puml 0.9.1')
    argparser.add_argument('path', metavar='path', type=str, help='the filepath to the domain')
    argparser.add_argument('module', metavar='module', type=str, help='the module name of the domain', default=None)

    args = argparser.parse_args()
    print(''.join(py2puml(args.path, args.module)))



from dataclasses import dataclass
from typing import List

from py2puml.domain.umlitem import UmlItem


@dataclass
class Member:
    name: str
    value: str


@dataclass
class UmlEnum(UmlItem):
    members: List[Member]



from dataclasses import dataclass
from typing import List

from py2puml.domain.umlitem import UmlItem


@dataclass
class UmlAttribute:
    name: str
    type: str
    static: bool


@dataclass
class UmlClass(UmlItem):
    attributes: List[UmlAttribute]
    is_abstract: bool = False



from io import StringIO
from pathlib import Path
from typing import Iterable, List, Union

from py2puml.py2puml import py2puml


def assert_py2puml_is_file_content(domain_path: str, domain_module: str, diagram_filepath: Union[str, Path]):
    # reads the existing class diagram
    with open(diagram_filepath, 'r', encoding='utf8') as expected_puml_file:
        assert_py2puml_is_stringio(domain_path, domain_module, expected_puml_file)


def normalize_lines_with_returns(lines_with_returns: Iterable[str]) -> List[str]:
    """
    When comparing contents, each piece of contents can either be:
    - a formatted string block output by the py2puml command containg line returns
    - a single line of contents read from a file, each line ending with a line return

    This function normalizes each sequence of contents as a list of string lines,
    each one finishing without a line return to ease comparison.
    """
    return ''.join(lines_with_returns).split('\n')


def assert_py2puml_is_stringio(domain_path: str, domain_module: str, expected_content_stream: StringIO):
    puml_content_lines = normalize_lines_with_returns(py2puml(domain_path, domain_module))
    expected_content_lines = normalize_lines_with_returns(expected_content_stream)

    assert_multilines(puml_content_lines, expected_content_lines)


def assert_multilines(actual_multilines: List[str], expected_multilines: List[str]):
    line_index = 0
    for line_index, (actual_line, expected_line) in enumerate(zip(actual_multilines, expected_multilines)):
        # print(f'{actual_line=}\n{expected_line=}')
        assert (
            actual_line == expected_line
        ), f'actual and expected contents have changed at line {line_index + 1}: {actual_line=}, {expected_line=}'

    assert line_index + 1 == len(actual_multilines), f'actual and expected diagrams have {line_index + 1} lines'



from importlib import import_module
from pkgutil import walk_packages
from types import ModuleType
from typing import Dict, List

from py2puml.domain.umlitem import UmlItem
from py2puml.domain.umlrelation import UmlRelation
from py2puml.inspection.inspectmodule import inspect_module


def inspect_package(
    domain_path: str, domain_module: str, domain_items_by_fqn: Dict[str, UmlItem], domain_relations: List[UmlRelation]
):
    # inspects the package module first, then its children modules and subpackages
    item_module = import_module(domain_module)
    inspect_module(item_module, domain_module, domain_items_by_fqn, domain_relations)

    for _, name, is_pkg in walk_packages([domain_path], f'{domain_module}.'):
        if not is_pkg:
            domain_item_module: ModuleType = import_module(name)
            inspect_module(domain_item_module, domain_module, domain_items_by_fqn, domain_relations)
    item_module = import_module(f'{domain_module}', f'{domain_module}.')
    inspect_module(item_module, domain_module, domain_items_by_fqn, domain_relations)



from enum import Enum
from typing import Dict, Type

from py2puml.domain.umlenum import Member, UmlEnum
from py2puml.domain.umlitem import UmlItem


def inspect_enum_type(enum_type: Type[Enum], enum_type_fqn: str, domain_items_by_fqn: Dict[str, UmlItem]):
    domain_items_by_fqn[enum_type_fqn] = UmlEnum(
        name=enum_type.__name__,
        fqn=enum_type_fqn,
        members=[Member(name=enum_member.name, value=enum_member.value) for enum_member in enum_type],
    )



from ast import AST, parse
from importlib import import_module
from inspect import getsource, unwrap
from textwrap import dedent
from typing import Dict, List, Tuple, Type

from py2puml.domain.umlclass import UmlAttribute
from py2puml.domain.umlrelation import UmlRelation
from py2puml.parsing.astvisitors import ConstructorVisitor
from py2puml.parsing.moduleresolver import ModuleResolver


def parse_class_constructor(
    class_type: Type, class_fqn: str, root_module_name: str
) -> Tuple[List[UmlAttribute], Dict[str, UmlRelation]]:
    constructor = getattr(class_type, '__init__', None)
    # conditions to meet in order to parse the AST of a constructor
    if (
        (  # the constructor must be defined
            constructor is None
        )
        or (  # the constructor's source code must be available
            not hasattr(constructor, '__code__')
        )
        or (  # the constructor must belong to the parsed class (not its parent's one)
            not constructor.__qualname__.endswith(f'{class_type.__name__}.__init__')
        )
    ):
        return [], {}

    # gets the original constructor, if wrapped by a decorator
    constructor = unwrap(constructor)

    constructor_source: str = dedent(getsource(constructor.__code__))
    constructor_ast: AST = parse(constructor_source)

    module_resolver = ModuleResolver(import_module(class_type.__module__))

    visitor = ConstructorVisitor(constructor_source, class_type.__name__, root_module_name, module_resolver)
    visitor.visit(constructor_ast)

    return visitor.uml_attributes, visitor.uml_relations_by_target_fqn



from typing import Dict, Iterable, List

from py2puml.domain.umlitem import UmlItem
from py2puml.domain.umlrelation import UmlRelation
from py2puml.export.puml import to_puml_content
from py2puml.inspection.inspectpackage import inspect_package


def py2puml(domain_path: str, domain_module: str) -> Iterable[str]:
    domain_items_by_fqn: Dict[str, UmlItem] = {}
    domain_relations: List[UmlRelation] = []
    inspect_package(domain_path, domain_module, domain_items_by_fqn, domain_relations)

    return to_puml_content(domain_module, domain_items_by_fqn.values(), domain_relations)



from typing import Iterable, List

from py2puml.domain.umlclass import UmlClass
from py2puml.domain.umlenum import UmlEnum
from py2puml.domain.umlitem import UmlItem
from py2puml.domain.umlrelation import UmlRelation

PUML_FILE_START = """@startuml {diagram_name}
!pragma useIntermediatePackages false

"""
PUML_FILE_FOOTER = """footer Generated by //py2puml//
"""
PUML_FILE_END = """@enduml
"""
PUML_ITEM_START_TPL = """{item_type} {item_fqn} {{
"""
PUML_ATTR_TPL = """  {attr_name}: {attr_type}{staticity}
"""
PUML_ITEM_END = """}
"""
PUML_RELATION_TPL = """{source_fqn} {rel_type}-- {target_fqn}
"""

FEATURE_STATIC = ' {static}'
FEATURE_INSTANCE = ''


def to_puml_content(diagram_name: str, uml_items: List[UmlItem], uml_relations: List[UmlRelation]) -> Iterable[str]:
    yield PUML_FILE_START.format(diagram_name=diagram_name)

    # exports the domain classes and enums
    for uml_item in uml_items:
        if isinstance(uml_item, UmlEnum):
            uml_enum: UmlEnum = uml_item
            yield PUML_ITEM_START_TPL.format(item_type='enum', item_fqn=uml_enum.fqn)
            for member in uml_enum.members:
                yield PUML_ATTR_TPL.format(attr_name=member.name, attr_type=member.value, staticity=FEATURE_STATIC)
            yield PUML_ITEM_END
        elif isinstance(uml_item, UmlClass):
            uml_class: UmlClass = uml_item
            yield PUML_ITEM_START_TPL.format(
                item_type='abstract class' if uml_item.is_abstract else 'class', item_fqn=uml_class.fqn
            )
            for uml_attr in uml_class.attributes:
                yield PUML_ATTR_TPL.format(
                    attr_name=uml_attr.name,
                    attr_type=uml_attr.type,
                    staticity=FEATURE_STATIC if uml_attr.static else FEATURE_INSTANCE,
                )
            yield PUML_ITEM_END
        else:
            raise TypeError(f'cannot process uml_item of type {uml_item.__class__}')

    # exports the domain relationships between classes and enums
    for uml_relation in uml_relations:
        yield PUML_RELATION_TPL.format(
            source_fqn=uml_relation.source_fqn, rel_type=uml_relation.type.value, target_fqn=uml_relation.target_fqn
        )

    yield PUML_FILE_FOOTER
    yield PUML_FILE_END



from ast import (
    AnnAssign,
    Assign,
    Attribute,
    BinOp,
    FunctionDef,
    Name,
    NodeVisitor,
    Subscript,
    arg,
    expr,
    get_source_segment,
)
from collections import namedtuple
from typing import Dict, List, Tuple

from py2puml.domain.umlclass import UmlAttribute
from py2puml.domain.umlrelation import RelType, UmlRelation
from py2puml.parsing.compoundtypesplitter import SPLITTING_CHARACTERS, CompoundTypeSplitter
from py2puml.parsing.moduleresolver import ModuleResolver

Variable = namedtuple('Variable', ['id', 'type_expr'])


class SignatureVariablesCollector(NodeVisitor):
    """
    Collects the variables and their type annotations from the signature of a constructor method
    """

    def __init__(self, constructor_source: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.constructor_source = constructor_source
        self.class_self_id: str = None
        self.variables: List[Variable] = []

    def visit_arg(self, node: arg):
        variable = Variable(node.arg, node.annotation)

        # first constructor variable is the name for the 'self' reference
        if self.class_self_id is None:
            self.class_self_id = variable.id
        # other arguments are constructor parameters
        else:
            self.variables.append(variable)


class AssignedVariablesCollector(NodeVisitor):
    """Parses the target of an assignment statement to detect whether the value is assigned to a variable or an instance attribute"""

    def __init__(self, class_self_id: str, annotation: expr):
        self.class_self_id: str = class_self_id
        self.annotation: expr = annotation
        self.variables: List[Variable] = []
        self.self_attributes: List[Variable] = []

    def visit_Name(self, node: Name):
        """
        Detects declarations of new variables
        """
        if node.id != self.class_self_id:
            self.variables.append(Variable(node.id, self.annotation))

    def visit_Attribute(self, node: Attribute):
        """
        Detects declarations of new attributes on 'self'
        """
        if isinstance(node.value, Name) and node.value.id == self.class_self_id:
            self.self_attributes.append(Variable(node.attr, self.annotation))

    def visit_Subscript(self, node: Subscript):
        """
        Assigns a value to a subscript of an existing variable: must be skipped
        """
        pass


class ConstructorVisitor(NodeVisitor):
    """
    Identifies the attributes (and infer their type) assigned to self in the body of a constructor method
    """

    def __init__(
        self, constructor_source: str, class_name: str, root_fqn: str, module_resolver: ModuleResolver, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.constructor_source = constructor_source
        self.class_fqn: str = f'{module_resolver.module.__name__}.{class_name}'
        self.root_fqn = root_fqn
        self.module_resolver = module_resolver
        self.class_self_id: str
        self.variables_namespace: List[Variable] = []
        self.uml_attributes: List[UmlAttribute] = []
        self.uml_relations_by_target_fqn: Dict[str, UmlRelation] = {}

    def extend_relations(self, target_fqns: List[str]):
        self.uml_relations_by_target_fqn.update(
            {
                target_fqn: UmlRelation(self.class_fqn, target_fqn, RelType.COMPOSITION)
                for target_fqn in target_fqns
                if target_fqn.startswith(self.root_fqn) and (target_fqn not in self.uml_relations_by_target_fqn)
            }
        )

    def get_from_namespace(self, variable_id: str) -> Variable:
        return next(
            (
                variable
                # variables namespace is iterated antichronologically
                # to account for variables being overridden
                for variable in self.variables_namespace[::-1]
                if variable.id == variable_id
            ),
            None,
        )

    def generic_visit(self, node):
        NodeVisitor.generic_visit(self, node)

    def visit_FunctionDef(self, node: FunctionDef):
        # retrieves constructor arguments ('self' reference and typed arguments)
        if node.name == '__init__':
            variables_collector = SignatureVariablesCollector(self.constructor_source)
            variables_collector.visit(node)
            self.class_self_id: str = variables_collector.class_self_id
            self.variables_namespace = variables_collector.variables

        self.generic_visit(node)

    def visit_AnnAssign(self, node: AnnAssign):
        variables_collector = AssignedVariablesCollector(self.class_self_id, node.annotation)
        variables_collector.visit(node.target)

        short_type, full_namespaced_definitions = self.derive_type_annotation_details(node.annotation)
        # if any, there is at most one self-assignment
        for variable in variables_collector.self_attributes:
            self.uml_attributes.append(UmlAttribute(variable.id, short_type, static=False))
            self.extend_relations(full_namespaced_definitions)

        # if any, there is at most one typed variable added to the scope
        self.variables_namespace.extend(variables_collector.variables)

    def visit_Assign(self, node: Assign):
        # recipients of the assignment
        for assigned_target in node.targets:
            variables_collector = AssignedVariablesCollector(self.class_self_id, None)
            variables_collector.visit(assigned_target)

            # attempts to infer attribute type when a single attribute is assigned to a variable
            if (len(variables_collector.self_attributes) == 1) and (isinstance(node.value, Name)):
                assigned_variable = self.get_from_namespace(node.value.id)
                if assigned_variable is not None:
                    short_type, full_namespaced_definitions = self.derive_type_annotation_details(
                        assigned_variable.type_expr
                    )
                    self.uml_attributes.append(
                        UmlAttribute(variables_collector.self_attributes[0].id, short_type, False)
                    )
                    self.extend_relations(full_namespaced_definitions)

            else:
                for variable in variables_collector.self_attributes:
                    short_type, full_namespaced_definitions = self.derive_type_annotation_details(variable.type_expr)
                    self.uml_attributes.append(UmlAttribute(variable.id, short_type, static=False))
                    self.extend_relations(full_namespaced_definitions)

            # other assignments were done in new variables that can shadow existing ones
            self.variables_namespace.extend(variables_collector.variables)

    def derive_type_annotation_details(self, annotation: expr) -> Tuple[str, List[str]]:
        """
        From a type annotation, derives:
        - a short version of the type (withenum.TimeUnit -> TimeUnit, Tuple[withenum.TimeUnit] -> Tuple[TimeUnit])
        - a list of the full-namespaced definitions involved in the type annotation (in order to build the relationships)
        """
        if annotation is None:
            return None, []

        # primitive type, object definition
        if isinstance(annotation, Name):
            full_namespaced_type, short_type = self.module_resolver.resolve_full_namespace_type(annotation.id)
            return short_type, [full_namespaced_type]
        # definition from module
        elif isinstance(annotation, Attribute):
            full_namespaced_type, short_type = self.module_resolver.resolve_full_namespace_type(
                get_source_segment(self.constructor_source, annotation)
            )
            return short_type, [full_namespaced_type]
        # compound type (List[...], Tuple[Dict[str, float], module.DomainType], etc.) or '|'-based union type
        elif isinstance(annotation, (Subscript, BinOp)):
            return shorten_compound_type_annotation(
                get_source_segment(self.constructor_source, annotation), self.module_resolver
            )

        return None, []


def shorten_compound_type_annotation(type_annotation: str, module_resolver: ModuleResolver) -> Tuple[str, List[str]]:
    """
    In the string representation of a compound type annotation, the elementary types can be prefixed by their packages or sub-packages.
    Like in 'Dict[datetime.datetime,typing.List[Worker]]'. This function returns a tuple of 2 values:
    - a string representation with shortened types for display purposes in the PlantUML documentation: 'Dict[datetime, List[Worker]]'
      (note: a space is inserted after each coma for readability sake)
    - a list of the fully-qualified types involved in the annotation: ['typing.Dict', 'datetime.datetime', 'typing.List', 'mymodule.Worker']
    """
    compound_type_parts: List[str] = CompoundTypeSplitter(type_annotation, module_resolver.module.__name__).get_parts()
    compound_short_type_parts: List[str] = []
    associated_types: List[str] = []
    for compound_type_part in compound_type_parts:
        # characters like '[', ']', ',', '|'
        if compound_type_part in SPLITTING_CHARACTERS:
            if compound_type_part == ',':
                compound_short_type_parts.append(', ')
            elif compound_type_part == '|':
                compound_short_type_parts.append(' | ')
            else:
                compound_short_type_parts.append(compound_type_part)
        # replaces each type definition by its short class name
        else:
            full_namespaced_type, short_type = module_resolver.resolve_full_namespace_type(compound_type_part)
            if short_type is None:
                raise ValueError(
                    f'Could not resolve type {compound_type_part} in module {module_resolver.module}: it needs to be imported explicitly.'
                )
            else:
                compound_short_type_parts.append(short_type)
            associated_types.append(full_namespaced_type)

    return ''.join(compound_short_type_parts), associated_types



from typing import Type


def investigate_domain_definition(type_to_inspect: Type):
    """
    Utilitary function which inspects the annotations of the given type
    """
    type_annotations = getattr(type_to_inspect, '__annotations__', None)
    if type_annotations is None:
        # print(f'class {type_to_inspect.__module__}.{type_to_inspect.__name__} of type {type(type_to_inspect)} has no annotation')
        for attr_class_key in dir(type_to_inspect):
            if attr_class_key != '__doc__':
                print(f'{type_to_inspect.__name__}.{attr_class_key}:', getattr(type_to_inspect, attr_class_key))
    else:
        # print(type_to_inspect.__annotations__)
        for attr_name, attr_class in type_annotations.items():
            for attr_class_key in dir(attr_class):
                if attr_class_key != '__doc__':
                    print(
                        f'{type_to_inspect.__name__}.{attr_name}:', attr_class_key, getattr(attr_class, attr_class_key)
                    )



from functools import reduce
from inspect import isclass
from types import ModuleType
from typing import Iterable, List, NamedTuple, Type


class NamespacedType(NamedTuple):
    """
    Information of a value type:
    - the module-prefixed type name
    - the short type name
    """

    full_namespace: str
    type_name: str


EMPTY_NAMESPACED_TYPE = NamespacedType(None, None)


def search_in_module_or_builtins(searched_module: ModuleType, namespace: str):
    if searched_module is None:
        return None

    # searches the namespace in the module definitions and imports
    found_module_or_leaf_type = getattr(searched_module, namespace, None)
    if found_module_or_leaf_type is not None:
        return found_module_or_leaf_type

    # searches the namespace in the builtins otherwise
    if hasattr(searched_module, '__builtins__'):
        return searched_module.__builtins__.get(namespace, None)


def search_in_module(namespaces: List[str], module: ModuleType):
    leaf_type: Type = reduce(search_in_module_or_builtins, namespaces, module)
    if leaf_type is None:
        return EMPTY_NAMESPACED_TYPE
    else:
        # https://bugs.python.org/issue34422#msg323772
        short_type = getattr(leaf_type, '__name__', getattr(leaf_type, '_name', None))
        return NamespacedType(f'{leaf_type.__module__}.{short_type}', short_type)


class ModuleResolver:
    """
    Given a module and a partially namespaced type name, returns a tuple of information about the type:
    - the full-namespaced type name, to draw relationships between types without ambiguity
    - the short type name, for display purposes

    Different approaches are combined to derive the type information:
    - when the partially namespaced type is found during AST parsing (class constructors)
    - when the partially namespaced type is found during class inspection (dataclasses, class static variables, named tuples, enums)

    The two approaches are a bit entangled for now, they could be separated a bit more for performance sake.
    """

    def __init__(self, module: ModuleType):
        self.module = module

    def __repr__(self) -> str:
        return f'ModuleResolver({self.module})'

    def resolve_full_namespace_type(self, partial_dotted_path: str) -> NamespacedType:
        """
        Returns a tuple of 2 strings:
        - the full namespaced type
        - the short named type
        """
        if partial_dotted_path is None:
            return EMPTY_NAMESPACED_TYPE

        # special case for Union types
        if partial_dotted_path == 'None':
            return NamespacedType('builtins.None', 'None')

        def string_repr(module_attribute) -> str:
            return (
                f'{module_attribute.__module__}.{module_attribute.__name__}'
                if isclass(module_attribute)
                else f'{module_attribute}'
            )

        # searches the class in the module imports
        namespaced_types_iter: Iterable[NamespacedType] = (
            NamespacedType(string_repr(getattr(self.module, module_var)), module_var)
            for module_var in vars(self.module)
        )
        found_namespaced_type = next(
            (
                namespaced_type
                for namespaced_type in namespaced_types_iter
                if namespaced_type.full_namespace == partial_dotted_path
            ),
            None,
        )

        # searches the class in the builtins
        if found_namespaced_type is None:
            found_namespaced_type = search_in_module(partial_dotted_path.split('.'), self.module)

        return found_namespaced_type

    def get_module_full_name(self) -> str:
        return self.module.__name__
